@@ -4,12 +4,17 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#if HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <assert.h>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <ctype.h>
 #include <errno.h>
 #include <netdb.h>
+#include <limits.h>
 #include <string.h>
 #include <strings.h>
 #include <stdarg.h>
@@ -302,7 +307,11 @@ int
 bgpq_expanded_v6prefix(char* prefix, struct bgpq_expander* ex,
 	struct bgpq_request* req)
 {
-	bgpq_expander_add_prefix(ex,prefix);
+	char* d = strchr(prefix, '^');
+	if (!d)
+		bgpq_expander_add_prefix(ex,prefix);
+	else
+		bgpq_expander_add_prefix_range(ex,prefix);
 	return 1;
 };
 
@@ -373,6 +382,29 @@ bgpq_pipeline(struct bgpq_expander* b,
 	return bp;
 };
 
+static void
+bgpq_expander_invalidate_asn(struct bgpq_expander* b, const char* q)
+{
+	if (!strncmp(q, "!gas", 4) || !strncmp(q, "!6as", 4)) {
+		char* eptr;
+		unsigned long asn = strtoul(q+4, &eptr, 10), asn0, asn1 = 0;
+		if (!asn || asn == ULONG_MAX || asn >= 4294967295 ||
+			(eptr && *eptr != '\n')) {
+			sx_report(SX_ERROR, "some problem invalidating asn %s\n", q);
+			return;
+		};
+		asn1 = asn % 65536;
+		asn0 = asn / 65536;
+		if (!b->asn32s[asn0] ||
+			!(b->asn32s[asn0][asn1/8] & (0x80 >> (asn1 % 8)))) {
+			sx_report(SX_NOTICE, "strange, invalidating inactive asn %lu(%s)\n",
+				asn, q);
+		} else {
+			b->asn32s[asn0][asn1/8] &= ~(0x80 >> (asn1 % 8));
+		};
+	};
+};
+
 static void
 bgpq_write(struct bgpq_expander* b)
 {
@@ -401,7 +433,7 @@ bgpq_selread(struct bgpq_expander* b, char* buffer, int size)
 {
 	fd_set rfd, wfd;
 	int ret;
-	struct timeval timeout = {10, 0};
+	struct timeval timeout = {30, 0};
 
 repeat:
 	FD_ZERO(&rfd);
@@ -465,6 +497,10 @@ have:
 			unsigned long togot=strtoul(response+1,&eon,10);
 			char* recvbuffer=malloc(togot+2);
 			int offset = 0;
+			if (!recvbuffer) {
+				sx_report(SX_FATAL, "error allocating %lu bytes: %s\n",
+					togot+2, strerror(errno));
+			};
 			memset(recvbuffer,0,togot+2);
 
 			if(!eon || *eon!='\n') {
@@ -550,10 +586,12 @@ have3:
 		} else if(response[0]=='C') {
 			/* No data */
 			SX_DEBUG(debug_expander,"No data expanding %s\n", req->request);
+			if (b->validate_asns) bgpq_expander_invalidate_asn(b, req->request);
 		} else if(response[0]=='D') {
 			/* .... */
 			SX_DEBUG(debug_expander,"Key not found expanding %s\n",
 				req->request);
+			if (b->validate_asns) bgpq_expander_invalidate_asn(b, req->request);
 		} else if(response[0]=='E') {
 			sx_report(SX_ERROR, "Multiple keys expanding %s: %s\n",
 				req->request, response);
@@ -623,8 +661,12 @@ repeat:
 	if(response[0]=='A') {
 		char* eon, *c;
 		long togot=strtoul(response+1,&eon,10);
-		char recvbuffer[togot+2];
+		char *recvbuffer = malloc(togot+2);
 		int  offset = 0;
+		if (!recvbuffer) {
+			sx_report(SX_FATAL, "Error allocating %lu bytes: %s\n",
+				togot+2, strerror(errno));
+		};
 
 		if(eon && *eon!='\n') {
 			sx_report(SX_ERROR,"A-code finised with wrong char '%c' (%s)\n",
@@ -689,10 +731,13 @@ have3:
 			c+=spn+1;
 		};
 		memset(recvbuffer, 0, togot+2);
+		free(recvbuffer);
 	} else if(response[0]=='C') {
 		/* no data */
+		if (b->validate_asns) bgpq_expander_invalidate_asn(b, request);
 	} else if(response[0]=='D') {
 		/* ... */
+		if (b->validate_asns) bgpq_expander_invalidate_asn(b, request);
 	} else if(response[0]=='E') {
 		/* XXXXXX */
 	} else if(response[0]=='F') {
@@ -711,6 +756,9 @@ bgpq_expand(struct bgpq_expander* b)
 	int fd=-1, err, ret;
 	struct sx_slentry* mc;
 	struct addrinfo hints, *res=NULL, *rp;
+	struct linger sl;
+	sl.l_onoff = 1;
+	sl.l_linger = 5;
 	memset(&hints,0,sizeof(struct addrinfo));
 
 	hints.ai_socktype=SOCK_STREAM;
@@ -730,6 +778,13 @@ bgpq_expand(struct bgpq_expander* b)
 				strerror(errno));
 			exit(1);
 		};
+		if (setsockopt(fd, SOL_SOCKET, SO_LINGER, &sl, sizeof(struct linger))) {
+			sx_report(SX_ERROR,"Unable to set linger on socket: %s\n",
+				strerror(errno));
+			shutdown(fd, SHUT_RDWR);
+			close(fd);
+			exit(1);
+		};
 		err=connect(fd,rp->ai_addr,rp->ai_addrlen);
 		if(err) {
 			shutdown(fd,SHUT_RDWR);
@@ -766,12 +821,15 @@ bgpq_expand(struct bgpq_expander* b)
 	};
 
 	if(b->sources && b->sources[0]!=0) {
-		char sources[128];
+		int slen = strlen(b->sources)+4;
+		if (slen < 128)
+			slen = 128;
+		char sources[slen];
 		snprintf(sources,sizeof(sources),"!s%s\n", b->sources);
 		SX_DEBUG(debug_expander,"Requesting sources %s", sources);
 		write(fd, sources, strlen(sources));
-		memset(sources, 0, sizeof(sources));
-		read(fd, sources, sizeof(sources));
+		memset(sources, 0, slen);
+		read(fd, sources, slen);
 		SX_DEBUG(debug_expander,"Got answer %s", sources);
 		if(sources[0]!='C') {
 			sx_report(SX_ERROR, "Invalid source(s) '%s': %s\n", b->sources,
@@ -812,8 +870,8 @@ bgpq_expand(struct bgpq_expander* b)
 			bgpq_read(b);
 	};
 
-	if(b->generation>=T_PREFIXLIST) {
-		unsigned i, j, k;
+	if(b->generation>=T_PREFIXLIST || b->validate_asns) {
+		uint32_t i, j, k;
 		STAILQ_FOREACH(mc, &b->rsets, next) {
 			if(b->family==AF_INET) {
 				bgpq_expand_irrd(b, bgpq_expanded_prefix, NULL, "!i%s,1\n",
@@ -830,35 +888,19 @@ bgpq_expand(struct bgpq_expander* b)
 					if(b->asn32s[k][i]&(0x80>>j)) {
 						if(b->family==AF_INET6) {
 							if(!pipelining) {
-								if(k>0)
-									bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
-										NULL, "!6as%u.%u\r\n", k, i*8+j);
-								else
-									bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
-										NULL,"!6as%u\r\n", i*8+j);
+								bgpq_expand_irrd(b, bgpq_expanded_v6prefix,
+									NULL, "!6as%" PRIu32 "\n", (k<<16)+i*8+j);
 							} else {
-								if(k>0)
-									bgpq_pipeline(b, bgpq_expanded_v6prefix,
-										NULL, "!6as%u.%u\r\n", k, i*8+j);
-								else
-									bgpq_pipeline(b,bgpq_expanded_v6prefix,
-										NULL, "!6as%u\r\n", i*8+j);
+								bgpq_pipeline(b, bgpq_expanded_v6prefix,
+									NULL, "!6as%" PRIu32 "\n", (k<<16)+i*8+j);
 							};
 						} else {
 							if(!pipelining) {
-								if(k>0)
-									bgpq_expand_irrd(b, bgpq_expanded_prefix,
-										NULL,"!gas%u.%u\n", k, i*8+j);
-								else
-									bgpq_expand_irrd(b, bgpq_expanded_prefix,
-										NULL,"!gas%u\n", i*8+j);
+								bgpq_expand_irrd(b, bgpq_expanded_prefix,
+									NULL, "!gas%" PRIu32 "\n", (k<<16)+i*8+j);
 							} else {
-								if(k>0)
-									bgpq_pipeline(b, bgpq_expanded_prefix,
-										NULL, "!gas%u.%u\n", k, i*8+j);
-								else
-									bgpq_pipeline(b, bgpq_expanded_prefix,
-										NULL, "!gas%u\n", i*8+j);
+								bgpq_pipeline(b, bgpq_expanded_prefix,
+									NULL, "!gas%" PRIu32 "\n", (k<<16)+i*8+j);
 							};
 						};
 					};
@@ -874,6 +916,11 @@ bgpq_expand(struct bgpq_expander* b)
 	};
 
 	write(fd, "!q\n",3);
+	if (pipelining) {
+		int fl = fcntl(fd, F_GETFL);
+		fl &= ~O_NONBLOCK;
+		fcntl(fd, F_SETFL, fl);
+	};
 	shutdown(fd, SHUT_RDWR);
 	close(fd);
 	return 1;
